package com.example.schemamapper;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Properties;
import java.io.*;

import org.apache.xerces.parsers.DOMParser;
import org.xml.sax.SAXException;
import org.w3c.dom.*;

/**
 * SchemaMapper Class
 * 
 * @author David Cleary
 * @version 1.0
 */
public class SchemaMapper
{
	// SchemaMapper static functions and members
	static Vector mappingObjects = null;	// Holds a SchemaMapper instance per Namespace
	static Properties schemas = null;		// Property file that maps Namespace to a Schema
	
	// This is the mapping vocabulary namespace
	static final String mappingURI = "http://www.example.com/schemamap";
	
	// This is the April 7 XML Schema namespace
	static final String schemaURI = "http://www.w3.org/1999/XMLSchema";
	
	/**
	 * The main processing method.
	 *
	 * @param nsURI Namespace URI of the root element.
	 * @param ele Root element of schema instance.
	 * @param xmlClass Object holder to return populated class
	 * @return boolean True if processInstance was successfull
	 */
	public static boolean processInstance(String nsURI, org.w3c.dom.Element ele, Object[] xmlClass)
	{
		int i;
		boolean ret = false;
		SchemaMapper mapper = null;
		
		if (mappingObjects != null)
		{
		
			// Walk through our vector looking for our namespace
			for (i = 0; i < mappingObjects.size(); i++)
			{
				mapper = (SchemaMapper)mappingObjects.elementAt(i);
				if (nsURI.compareTo(mapper.getNamespaceURI()) == 0)
					break;
				else
					mapper = null;
			}
		}
		
		if (mapper == null)
		{
			mapper = SchemaMapper.createMappingObject(nsURI);
		}
			
		
		if (mapper != null)
		{
			xmlClass[0] = mapper.mapClass(ele);
			ret = true;
		}
			
				
		return ret;
	}
	
	/**
	 * A Properties file breaks with : in the property name. This
	 * routing replaces : with ? in namespace URIs.
	 *
	 * @param nsURI Namespace URI to encode.
	 */
	public static String encodeURI(String nsURI)
	{
		int i;
		StringBuffer retString = new StringBuffer(nsURI);
		
		for (i = 0; i < nsURI.length(); i++)
		{
			char x = retString.charAt(i);
			if (x == ':')
				retString.setCharAt(i, '?');
		}
		
		return new String(retString);
	}
	
	/**
	 * This routine parses a schema and creates a SchemaMapper
	 * object that contains all the info to map a particular
	 * namespace.
	 *
	 * @param namespaceURI Namespace URI to encode.
	 * @return SchemaMapper If successfull, a SchemaMapper object.
	 */
	public static SchemaMapper createMappingObject(String namespaceURI)
	{
		String schemaName;
		SchemaMapper map = null;
		
		// Create the mappingObjects vector if not done.
		if (mappingObjects == null)
			mappingObjects = new Vector();
		
		// schemas is the property file that relates a namespace to a
		// physical schema.
		if (schemas == null)
		{
			schemas = new Properties();
			// Load our schema prop file
			try
			{
				FileInputStream in = new FileInputStream("./Schemas/Schema.dat");
				schemas.load(in);
				in.close();
			}
			catch (IOException ex)
			{
				ex.printStackTrace();
			}
		}

		// Using the namespace, find the schema file.
		schemaName = schemas.getProperty(SchemaMapper.encodeURI(namespaceURI));
		if (schemaName != null)
		{
			// We need to walk the schema.
			DOMParser parser = new DOMParser();
			try
			{
				parser.parse("./Schemas/" + schemaName);

				// Create a new SchemaMapper for this namespace and add to the
				// vector.
				org.w3c.dom.Document xdoc = parser.getDocument();
				map = new SchemaMapper();
				map.parseSchema(namespaceURI, xdoc);
				if (map.valid)
					mappingObjects.addElement(map);
				else
					map = null;

			} 
			catch (SAXException se) 
			{
			    se.printStackTrace();
			} 
			catch (IOException ioe) {
			    ioe.printStackTrace();
			}
		}
				
		return map;
	}
	
	// SchemaMapper instance functions and members
	
	String namespaceURI;
	String className;
	String rootName;
	Hashtable mapInfoObjects;
	boolean valid;
	
	SchemaMapper()
	{
		namespaceURI = null;
		className = null;
		rootName = null;
		mapInfoObjects = new Hashtable();
		valid = false;
	}
	
	/**
	 * This routine parses a schema and creates MapInfo objects
	 * for each node that is mapped.
	 *
	 * @param nsURI Namespace URI of instance document.
	 * @param xdoc Document object of XML Schema.
	 */
	private void parseSchema(String nsURI, org.w3c.dom.Document xdoc)
	{
		int i;
		String nsPrefix;
		String path;
		boolean annotated = false;
		
		namespaceURI = nsURI;

		// Confirm we have mapping info by looking at xmlns attributes
		org.w3c.dom.Element root = xdoc.getDocumentElement();
		org.w3c.dom.NamedNodeMap attrMap = root.getAttributes();
		
		// Walk the NamedNodeMap because the prefix used is unimportant
		// and unknown.
		for (i = 0; i < attrMap.getLength(); i++)
		{
			org.w3c.dom.Attr nsAttr = (org.w3c.dom.Attr)attrMap.item(i);
			nsPrefix = nsAttr.getPrefix();
			if ((nsPrefix != null) && (nsPrefix.compareTo("xmlns") == 0))
			{
				String URI = nsAttr.getNodeValue();
				if (URI.compareTo(SchemaMapper.mappingURI) == 0)
				{
					annotated = true;
					break;
				}
			}
		}
		
		// We have the NS decl, so continue
		if (annotated)
		{
			// Look for a map:class attribute on the first element child off 
			// the root. A robust implementation would support multiple classes
			// and not be bound to the simple schemas this code supports.
			org.w3c.dom.Attr attr;
			org.w3c.dom.Node node;
			org.w3c.dom.Element ele = null;
			org.w3c.dom.NodeList list = root.getChildNodes();
			for (i = 0; i < list.getLength(); i++)
			{
				node = list.item(i);
				if (node.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE)
				{
					ele = (org.w3c.dom.Element)node;
					break;
				}
			}
			if (ele != null)
			{
				attrMap = ele.getAttributes();
				attr = (org.w3c.dom.Attr)attrMap.getNamedItemNS(mappingURI, "class");
				className = attr.getValue();
				attr = (org.w3c.dom.Attr)attrMap.getNamedItemNS(null, "name");
				rootName = attr.getValue();
				path = "/" + rootName;
				
				// walkSchema is called recursively as the schema is walked.
				valid = this.walkSchema(path, ele);
			}
		}
		
		
	}
	
	boolean walkSchema(String path, org.w3c.dom.Element ele)
	{
		int i;
		String xPath = path;
		org.w3c.dom.Node node;
		org.w3c.dom.Attr attr;
		org.w3c.dom.NamedNodeMap attrMap;
		org.w3c.dom.NodeList list = ele.getChildNodes();
		
		for (i = 0; i < list.getLength(); i++)
		{
			node = list.item(i);
			if (node.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE)
			{
				String nsURI = node.getNamespaceURI();
				String nodeName = node.getLocalName();
				
				if ((nsURI != null) && (nsURI.compareTo(schemaURI) == 0))
				{
					// No support for attributes in this code.
					if (nodeName.compareTo("element") == 0)
					{
						attrMap = node.getAttributes();
						attr = (org.w3c.dom.Attr)attrMap.getNamedItemNS(null, "name");
						if (attr != null)
						{
							nodeName = attr.getValue();
							xPath = path + "/" + nodeName;
						}
						
						attr = (org.w3c.dom.Attr)attrMap.getNamedItemNS(mappingURI, "name");
						if (attr != null)
						{
							MapInfo mi = new MapInfo();
							mi.setPath(xPath);
							mi.setName(attr.getValue());
							attr = (org.w3c.dom.Attr)attrMap.getNamedItemNS(mappingURI, "type");
							mi.setType(attr.getValue());
							attr = (org.w3c.dom.Attr)attrMap.getNamedItemNS(null, "type");			
							mi.setSchemaType(attr.getValue());
							this.mapInfoObjects.put(new String(xPath), mi);
						}
					}
				}
				this.walkSchema(xPath, (org.w3c.dom.Element)node);
			}
		}
		
		return true;
	}
	
	private Object mapClass(org.w3c.dom.Element ele)
	{
		String xPath;
		Object xmlClass = null;
		// This should use introspection
		if (this.className.compareTo("com.example.subscriber") == 0)
			xmlClass = new com.example.Subscriber();
		
		xPath = "";
		// Using XPath would be preferable to walking instance
		walkInstance(xPath, ele, (com.example.Subscriber)xmlClass);
		return xmlClass;
	}
	
	private void walkInstance(String xPath, org.w3c.dom.Element ele, com.example.Subscriber xmlClass)
	{
		int i;
		String eleName;
		MapInfo mappingInfo;
		org.w3c.dom.NodeList children;
		
		eleName = ele.getLocalName();
		xPath = xPath + "/" + eleName;
		mappingInfo = (MapInfo)mapInfoObjects.get(xPath);
		if (mappingInfo != null)
		{
			populateSubscriber(ele, mappingInfo, xmlClass);
		}
		
		children = ele.getChildNodes();
		
		for (i = 0; i < children.getLength(); i++)
		{
			org.w3c.dom.Node child = children.item(i);
			if (child.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE)
			{
				walkInstance(xPath, (org.w3c.dom.Element)child, xmlClass);
			}
		}
	}
	
	void populateSubscriber(org.w3c.dom.Element ele, MapInfo mappingInfo, com.example.Subscriber xmlClass)
	{
		String name = mappingInfo.getName();
		String value = getElementValue(ele);
		if (name.compareTo("FirstName") == 0)
		{
			if (mappingInfo.getType() == MapInfo.MI_PROPERTY)
				xmlClass.setFirstName(value);
		}
		else if (name.compareTo("LastName") == 0)
		{
			if (mappingInfo.getType() == MapInfo.MI_PROPERTY)
				xmlClass.setLastName(value);
		}
		else if (name.compareTo("BirthDate") == 0)
		{
			if (mappingInfo.getType() == MapInfo.MI_PROPERTY)
				xmlClass.setBirthDate(value);
		}
		else if (name.compareTo("Phone") == 0)
		{
			if (mappingInfo.getType() == MapInfo.MI_PROPERTY)
				xmlClass.setPhoneNumber(value);
		}
		else if (name.compareTo("splitName") == 0)
		{
			if (mappingInfo.getType() == MapInfo.MI_METHOD)
				xmlClass.splitName(value);
		}
	}
	
	private String getElementValue(org.w3c.dom.Element ele)
	{
		int i;
		String retVal = null;
		org.w3c.dom.NodeList list = ele.getChildNodes();
		for (i = 0; i < list.getLength(); i++)
		{
			org.w3c.dom.Node node = list.item(i);
			if (node.getNodeType() == org.w3c.dom.Node.TEXT_NODE)
			{
				retVal = node.getNodeValue();
				break;
			}
		}
		
		return retVal;
	}
				
	
	String getNamespaceURI()
	{
		return namespaceURI;
	}
	
	void setNamespaceURI(String newURI)
	{
		namespaceURI = newURI;
	}
}
